/*
 * s390-tools/tunedasd/include/tunedasd.h
 *   zSeries DASD tuning program.
 *
 * (C) Copyright IBM Corp. 2004
 *
 * Author(s): Horst Hummel (horst.hummel@de.ibm.com)
 */

#ifndef TUNEDASD_H
#define TUNEDASD_H

#include <stdint.h>
#include <unistd.h>
#include <getopt.h>

#define STRINGIFY_1(x)			#x
#define STRINGIFY(x)			STRINGIFY_1(x)

#define RELEASE_STRING	STRINGIFY (S390_TOOLS_RELEASE)

void error_print (const char* fmt, ...);

#endif /* not TUNEDASD_H */
