/*
 *  evaluate_xml.h
 *
 *    Copyright (C) 2003 IBM Deutschland Entwicklung GmbH, IBM Corporation
 */


#ifndef evaluate_xml_h
#define evaluate_xml_h

#include "xml.h"

#ifdef __s390x__
#define FMT64 "l"
#else
#define FMT64 "ll"
#endif

typedef struct   /* info given in a fcp_ipl section of XML */
{
	int                devno;
	unsigned long long wwpn;
	unsigned long long fcp_lun;
	unsigned int       configuration;
	char	           *type;
} fcp_info_t;

int get_fcp_info (char* ldipl_xml, fcp_info_t* fcp_info);
#endif
