/*
 *  drivers/s390/char/dump.h
 *
 *    Copyright (C) 2003 IBM Deutschland Entwicklung GmbH, IBM Corporation
 *    Author(s): Michael Holzheu <holzheu@de.ibm.com>
 */

#define ZCORE_IOCTL_XMLSIZE     0x1
#define ZCORE_IOCTL_GETXML      0x2
#define ZCORE_IOCTL_RELEASE_HSA 0x3
#define ZCORE_IOCTL_GET_ERROR   0x4
#define ZCORE_IOCTL_HSASIZE     0x5

#define ZCORE_ERR_OK            0x0
#define ZCORE_ERR_HSA_FEATURE   0x1
#define ZCORE_ERR_CPU_INFO      0x2
#define ZCORE_ERR_OTHER         0x3

void setup_dump(char* cmdline);
int  dump_init(void);
